from __future__ import print_function

from base64 import b32encode
from binascii import unhexlify
import hashlib
import os
from typing import Union, Dict, AnyStr

from pyotp import TOTP
import xmltodict


__all__ = ['get_authenticators']


def _get_totp_obj(name: str, secret: AnyStr, digits: Union[str, int] = 6, digest: str = 'sha1', interval: Union[str, int] = 30):
    return TOTP(
        name=name,
        s=b32encode(unhexlify(secret)),
        digits=int(digits),
        digest=getattr(hashlib, digest.lower()),
        interval=int(interval),
    )


def _normalize_name(name: str) -> str:
    return '_'.join(name.lower().split())


def get_authenticators(xml_path: Union[str, None] = None) -> Dict[str, TOTP]:
    if not xml_path:
        xml_path = os.path.join(os.environ['APPDATA'], 'WinAuth', 'winauth.xml')
    with open(xml_path, 'rb') as f:
        xml_data = xmltodict.parse(f, force_list=['WinAuthAuthenticator'])
        if 'WinAuthAuthenticator' not in xml_data['WinAuth']:
            return {}
        authenticators = {_normalize_name(i['name']):
                          _get_totp_obj(i['name'], *i['authenticatordata']['secretdata'].split('\t'))
                          for i in xml_data['WinAuth']['WinAuthAuthenticator']
                          if i['@type'] == 'WinAuth.GoogleAuthenticator'}
    return authenticators



from __future__ import print_function

import argparse
import sys
import warnings

from . import get_authenticators

try:
    import pyperclip
    have_pyperclip = True
except ImportError:
    have_pyperclip = False

try:
    warnings.filterwarnings('ignore', module='fuzzywuzzy')
    from fuzzywuzzy import process
    from fuzzywuzzy.fuzz import QRatio
    have_fuzzy = True
except ImportError:
    have_fuzzy = False

try:
    import qrcode
    have_qrcode = True
except ImportError:
    have_qrcode = False


def _output(s: str) -> None:
    if not s:
        return
    print(s)
    if have_pyperclip:
        try:
            pyperclip.copy(s)
        except Exception as e:
            print('Clipboard failed! {}'.format(e), file=sys.stderr)


def main() -> int:  # pylint: disable=inconsistent-return-statements
    p = argparse.ArgumentParser()
    p.add_argument('-x', '--xml-path')
    p.add_argument('--secret', dest='actions', action='append_const', const='secret')
    p.add_argument('--url', dest='actions', action='append_const', const='url')
    if have_qrcode:
        p.add_argument('--qr', dest='actions', action='append_const', const='qr')
    p.add_argument('name', nargs='?')
    args = p.parse_args()

    try:
        authenticators = get_authenticators(args.xml_path)
    except Exception as e:
        print('Could not load WinAuth configuration! {}'.format(e), file=sys.stderr)
        return 2

    if args.name is None:
        for name in sorted(authenticators):
            print(name)
        return 1

    if have_fuzzy:
        name, certainty, auth = process.extractOne(args.name, {v: k for k, v in authenticators.items()}, scorer=QRatio)
        if certainty < 15:
            print('Authenticator {} not found!'.format(sys.argv[1]), file=sys.stderr)
            return 1
        if certainty < 100:
            print('Guessed authenticator {}'.format(name), file=sys.stderr)
    else:
        try:
            auth = authenticators[args.name]
        except KeyError:
            print('Authenticator {} not found!'.format(sys.argv[1]), file=sys.stderr)
            return 1

    if not args.actions:
        _output(auth.now())
        return

    if 'secret' in args.actions:
        _output(auth.secret.decode())

    if 'url' in args.actions:
        _output(auth.provisioning_uri())

    if 'qr' in args.actions:
        qr = qrcode.QRCode()
        qr.add_data(auth.provisioning_uri())
        qr.print_ascii(invert=True)


if __name__ == '__main__':
    sys.exit(main())



from setuptools import setup


setup(
    name='pywinauth',
    version='0.4.0',
    url='https://github.com/zeevro/pywinauth',
    download_url='https://github.com/zeevro/pywinauth/archive/master.zip',
    author='Zeev Rotshtein',
    author_email='zeevro@gmail.com',
    maintainer='Zeev Rotshtein',
    maintainer_email='zeevro@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: Public Domain',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Security',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
    license=None,
    description='A tool to generate TOTP code based using secrets from WinAuth',
    keywords=[
        'WinAuth',
        'TOTP',
        '2FA',
        'MFA',
    ],
    zip_safe=True,
    packages=[
        'winauth',
    ],
    install_requires=[
        'xmltodict',
        'pyotp',
        'pyperclip',
        'fuzzywuzzy',
        'qrcode',
    ],
    entry_points=dict(
        console_scripts=[
            'winauthc = winauth.__main__:main',
        ],
    ),
)
